import cv2
from cv2.typing import MatLike

import numpy as np


class Autograph():
    """

    Класс для получения подписи из документа (pdf, png, jpg, bmp)
        - color_low - нижняя граница hsv цвета, по которому находится цветная печать и подпись, по стандарту [0, 50, 0]
        - color_hight верхняя граница hsv цвета, по которому находится цветная печать и подпись, по стандарту [255, 255, 255]
        - blur - насколько сильно будет размываться изображение, чем сильнее размытие, тем сильнее печать становиться более круглой, что упростит ее нахождение и удаление, по стандарту (3, 3)
        - min_radius - минимальный радиус для удаления окружности (печати), по стандарту 80
        - max_radius - максимальный радиус для удаления окружности (печати), по стандарту 200
        - precent_expansion - увеличение области удаления окружности (печати), чем больше значение, тем более большой круг вырежеться, если 0 - то будет окружность будет вырезана четко по контуру, оставляя небольшие следы из пикселей, по стандарту 0.15
        - pixel_thickness - ширина пикселей, чем больше - тем меньше шанс, что подпись разорвертся, но при этом у выходного изображения будут широкие пиксели
        - size - размер выходного изображения, по стандарту (256, 256)

    """

    def __init__(
            self,
            color_low=[0, 50, 0],
            color_hight=[255, 255, 255],
            blur=(3, 3),
            min_radius=80,
            max_radius=200,
            precent_expansion=0.15,
            pixel_thickness=(3, 3),
            size=(256, 256)
    ) -> None:
        self.__color_low = color_low
        self.__color_hight = color_hight
        self.__blur = blur
        self.__min_radius = min_radius
        self.__max_radius = max_radius
        self.__precent_expansion = precent_expansion
        self.__pixel_thickness = pixel_thickness
        self.__size = size

    def __remove_text(self, picture: MatLike) -> MatLike:
        """
        Метод для очистки документа от текста, оставляя только печать и подпись в заданом диапозоне цветов, по стандарту оставляет синий цвет
            - picture - объект MatLike из opencv

        Возвращает объект MatLike из opencv
        """
        hsv = cv2.cvtColor(picture, cv2.COLOR_BGR2HSV)

        low = np.array(self.__color_low)
        hight = np.array(self.__color_hight)

        mask = cv2.inRange(hsv, low, hight)
        picture = (255 - mask)

        rgb = cv2.cvtColor(picture, cv2.COLOR_RGB2BGR)
        return rgb

    def __remove_print(self, picture: MatLike) -> MatLike:
        """
        Метод для удаления круглой печати
            - picture - объект MatLike из opencv

        Возвращает объект MatLike из opencv
        """
        gray_blurred = cv2.blur(picture, self.__blur)
        detected_circles = cv2.HoughCircles(
            image=gray_blurred,
            method=cv2.HOUGH_GRADIENT,
            dp=1,
            minDist=500,
            param1=50,
            param2=30,
            minRadius=self.__min_radius,
            maxRadius=self.__max_radius
        )

        if detected_circles is None:
            return picture

        detected_circles = np.uint16(np.around(detected_circles))

        first_point = detected_circles[0, :][0]
        a, b, r = first_point[0], first_point[1], first_point[2]

        cv2.circle(
            picture,
            (a, b),
            int(r + (r * self.__precent_expansion)),
            (255, 255, 255),
            -1
        )

        return picture

    def __finishing_lines(self, picture: MatLike) -> MatLike:
        """
        Метод для удаление разрывов на подписи, так как при удалении текста и печать, в некоторых местах она разрывается из-за чего в последствии она обрежеться не корректно
            - picture - объект MatLike из opencv

        Возвращает объект MatLike из opencv
        """
        kernel = np.ones(self.__pixel_thickness, np.uint8)
        picture = cv2.erode(picture, kernel, iterations=1)
        return picture

    def __skeletonization(self, picture: MatLike) -> MatLike:
        """
        Метод для скелетирования изображения (метдо работает плохо, из-за этого не используется)
            - picture - объект MatLike из opencv

        Возвращает объект MatLike из opencv
        """
        picture = cv2.bitwise_not(picture)

        size = np.size(picture)
        skeleton = np.zeros(picture.shape, np.uint8)

        element = cv2.getStructuringElement(cv2.MORPH_CROSS, (3, 3))
        done = False

        while (not done):
            eroded = cv2.erode(picture, element)
            temp = cv2.dilate(eroded, element)
            temp = cv2.subtract(picture, temp)
            skeleton = cv2.bitwise_or(skeleton, temp)
            picture = eroded.copy()

            zeros = size - cv2.countNonZero(picture)
            if zeros == size:
                done = True

        return skeleton

    def __crop_picture(self, picture: MatLike) -> MatLike:
        """
        Метод для обрезки изображения по контурам подписи
            - picture - объект MatLike из opencv

        Возвращает объект MatLike из opencv
        """
        _, thresh_gray = cv2.threshold(
            picture,
            thresh=100,
            maxval=255,
            type=cv2.THRESH_BINARY_INV
        )

        contours, _ = cv2.findContours(
            thresh_gray,
            cv2.RETR_LIST,
            cv2.CHAIN_APPROX_SIMPLE
        )

        max_box = (0, 0, 0, 0)
        max_area = 0
        for cont in contours:
            x, y, w, h = cv2.boundingRect(cont)
            area = w * h
            if area > max_area:
                max_box = x, y, w, h
                max_area = area
        x, y, w, h = max_box

        crop_picture = picture[y:y+h, x:x+w]
        return crop_picture

    def resize_picture(self, picture: MatLike) -> MatLike:
        """
        Метод для изменеия размеров изображения
            - picture - объект MatLike из opencv

        Возвращает объект MatLike из opencv
        """
        picture = cv2.resize(picture, self.__size)
        return picture

    def rotate_picture(self, picture: MatLike, direction: int) -> MatLike:
        """
        Метод для переворачивания изображения
            - picture - объект MatLike из opencv
            - direction - rotateCode

        Возвращает объект MatLike из opencv
        """
        picture = cv2.rotate(picture, direction)
        return picture

    def get_clear_autograph(self, path: str) -> MatLike:
        """
        Метод для получения готовой, очищенной подписи
            - path - путь к изображению

        Возвращает объект MatLike из opencv
        """
        picture = cv2.imdecode(np.fromfile(path, dtype=np.uint8), cv2.IMREAD_COLOR)

        picture = self.__remove_text(picture)

        picture = cv2.cvtColor(picture, cv2.COLOR_BGR2GRAY)

        picture = self.__remove_print(picture)
        picture = self.__finishing_lines(picture)
        picture = self.__crop_picture(picture)
        # picture = self.__skeletonization(picture)
        picture = self.resize_picture(picture)

        return picture



class ProgressBar():
    """

    Класс для создания прогресс бара

    Взят от сюда: https://stackoverflow.com/questions/3173320/text-progress-bar-in-terminal-with-block-characters
    
    """
    
    def print(iteration, total, prefix='', suffix='', decimals=1, length=25, fill='█', printEnd="\r"):
        percent = ("{0:." + str(decimals) + "f}").format(100 * (iteration / float(total)))
        filledLength = int(length * iteration // total)
        bar = fill * filledLength + '-' * (length - filledLength)
        
        print(f'\r{prefix} |{bar}| {percent}% {suffix}', end=printEnd)

        # Print New Line on Complete
        if iteration == total:
            print()



import os
import cv2
import fitz
import shutil
import unidecode

from inskrib.autograph import Autograph
from inskrib.utils import ProgressBar


class Document():
    """

    Класс для обработки большого количества документов 
        - result_path - путь для сохранения всех обработанных файлов, по стандарту 'result',
        - result_autographs - путь для сохранения готовых подписей, по стандарту "result/autographs",
        - result_persons - путь для сохранения индексированных людей которым принадлежит подпись, по стандарту "result/persons.csv",
        - result_filenames - путь для сохранения, по стандарту "result/filenames.csv",
        - result_temp - путь для сохранения временных файлов, по стандарту "result/temp",
        - output_picture_type - формат сохранения файлов, по стандарту "png",
        - grouping - будут ли подписи группироваться по человеку, по стандарту False

    """

    def __init__(
        self,
        result_path='result',
        result_autographs="result/autographs",
        result_persons="result/persons.csv",
        result_filenames="result/filenames.csv",
        result_temp="result/temp",
        output_picture_type="png",
        grouping: bool = False
    ) -> None:
        self.__result_path = result_path
        self.__result_autographs = result_autographs
        self.__result_persons = result_persons
        self.__result_filenames = result_filenames
        self.__result_temp = result_temp
        self.__output_picture_type = output_picture_type
        self.__grouping = grouping

    def __create_storage(self) -> None:
        """
        Метод для создания всех нужных директорий
        """
        if not os.path.exists(self.__result_path):
            os.mkdir(self.__result_path)
        if not os.path.exists(self.__result_autographs):
            os.mkdir(self.__result_autographs)
        if not os.path.exists(self.__result_temp):
            os.mkdir(self.__result_temp)

        open(self.__result_filenames, "w")
        open(self.__result_persons, "w")

    def __write_new_person(self, person: str) -> None:
        """
        Метод для записи всех людей в отдельный csv файл
            - person - имя человека
        """
        with open(self.__result_persons, 'a') as file:
            file.write(f'{person}\n')

    def __write_new_filename(self, filename: str) -> None:
        """
        Метод для записи имен файлов всех сохраненных подписей в отдельный csv файл
            - filename - имя файла
        """
        with open(self.__result_filenames, 'a') as file:
            file.write(f'{filename}\n')

    def __get_person_name(self, dirpath: str) -> str:
        """
        Метод для получения имени человека в транслите
            - dirpath - полный путь до директории этого человека

        Возвращает строку с обработанным именем
        """
        person = dirpath.split('\\').pop()
        person = unidecode.unidecode(person)
        return person

    def __pdf_to_image(self, path_to_file: str, path_to_save: str):
        """
        Метод для перевода pdf файла в изображение
            - path_to_file - путь до pdf файла
            - path_to_save - пусть для сохранения готового изображения
        """
        with fitz.open(path_to_file) as pdf:
            page = pdf.load_page(0)
            pix = page.get_pixmap()
            pix.save(path_to_save)

    def __save_temp_file(self, dirpath: str, filename: str, person: str, id: str, index: str) -> None:
        """
        Метод для перевода pdf файла в изображение
            - dirpath - олный путь до директории этого человека 
            - filename - имя файла
            - person - имя человека
            - id - индетификатор человека
            - index - порядковый номер файла
        """
        splited_filename = filename.split('.')

        path_to_file = f'{dirpath}/{filename}'
        new_file_name = f'{id}-{person}-{index}.{self.__output_picture_type}'
        path_to_save = f'{self.__result_temp}/{new_file_name}'

        if (splited_filename.pop() == 'pdf'):
            self.__pdf_to_image(path_to_file, path_to_save)
            return

        shutil.copyfile(path_to_file, path_to_save)

    def __save_authograph(self, path, picture) -> None:
        """
        Метод для сохранения подписи
            - path - путь сохранения подписи
            - picture - изображение в виде MatLike из opencv
        """
        cv2.imwrite(path, picture)

    def __process_temp(self, path: str) -> None:
        """
        Метод для первоначальной обработки документов
            - path - путь до директории с документами, которые нужно обработать
        """
        prefix = 'Process Temp Files:'
        length = len(os.listdir(path)) - 2
        ProgressBar.print(0, length, prefix)

        id = 0
        for dirpath, _, filenames in os.walk(path):
            person = self.__get_person_name(dirpath)

            if (person == path):
                continue

            grouping_path = f'{self.__result_autographs}/{person}'
            if self.__grouping and not os.path.exists(grouping_path):
                os.mkdir(grouping_path)

            self.__write_new_person(f'{person},{id}')

            index = 0
            for filename in filenames:
                self.__save_temp_file(dirpath, filename, person, id, index)
                index += 1

            id += 1
            ProgressBar.print(id, length, prefix)

    def __process_authographs(self, autograph: Autograph) -> None:
        """
        Метод для получения подписей из всех документов
            - autograph - инстанс класса inskrib.Autograph
        """
        prefix = 'Process Authograph:'
        length = len(os.listdir(self.__result_temp))
        ProgressBar.print(0, length, prefix)

        index = 0
        for picture in os.listdir(self.__result_temp):
            path_to_save = f'{self.__result_autographs}/{picture}'
            path_to_picture = f'{self.__result_temp}/{picture}'

            if (self.__grouping):
                person = picture.split('-')[1]
                path_to_save = f'{self.__result_autographs}/{person}/{picture}'

            id = picture.split('-')[0]

            try:
                ath = autograph.get_clear_autograph(path_to_picture)
                self.__save_authograph(path_to_save, ath)
                self.__write_new_filename(f'{picture},{id}')
            except Exception:
                pass

            index += 1
            ProgressBar.print(index, length, prefix)

    def __remove_temp(self) -> None:
        """
        Метод для удаления temp (временной) директории
        """
        shutil.rmtree(self.__result_temp)

    def set_grouping(self, grouping_type: bool) -> None:
        """
        Метод для изменения состояния у переменной __grouping
        """
        self.__grouping = grouping_type

    def get_authoraphs(self, path: str, autograph: Autograph, remove_temp: bool = True) -> None:
        """
        Метод для получения готовых подписей и csv файлов из документов
            - path - путь до директории с документами
            - autograph - инстанс класса inskrib.Autograph
            - remove_temp - отвечает за удаление temp (временной) директории, по стандарту True 
        """
        self.__create_storage()
        self.__process_temp(path)
        self.__process_authographs(autograph)

        if remove_temp:
            self.__remove_temp()
            print('Temp Directory Removed')



from setuptools import setup, find_packages
from pathlib import Path


"""
:authors: Elisha Flacon
:license: BSD 3-Clause License, see LICENSE file
:copyright: (c) 2023 Elisha Flacon
"""


VERSION = '0.0.1.4'
NAME = 'inskrib'

HERE = Path(__file__).parent.resolve()

AUTHOR = 'Elisha Flacon'
DESCRIPTION = 'Inskrib - simple library for getting signature from docs'
LONG_DESCRIPTION = Path(HERE, 'README.md').read_text(encoding='utf-8')

URL = 'https://github.com/ElishaFlacon/inskrib'
DOWNLOAD_URL = 'https://github.com/ElishaFlacon/inskrib/archive/main.zip'

REQUIRES_PYTHON = '>=3.8'
LICENSE = 'BSD 3-Clause'


setup(
    name=NAME,
    version=VERSION,
    author=AUTHOR,

    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type='text/markdown',

    url=URL,
    download_url=DOWNLOAD_URL,

    python_requires=REQUIRES_PYTHON,

    license=LICENSE,

    packages=find_packages(),
    include_package_data=True,
    install_requires=['opencv-python', 'unidecode', 'PyMuPDF'],

    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ]
)
